@@ -39,6 +39,7 @@ import org.jetbrains.kotlin.com.intellij.openapi.Disposable
 import org.jetbrains.kotlin.com.intellij.openapi.project.Project
 import org.jetbrains.kotlin.com.intellij.openapi.util.Disposer.dispose
 import org.jetbrains.kotlin.com.intellij.openapi.util.Disposer.newDisposable
+import org.jetbrains.kotlin.compiler.plugin.ComponentRegistrar
 
 import org.jetbrains.kotlin.config.AnalysisFlags
 import org.jetbrains.kotlin.config.ApiVersion
@@ -60,6 +61,7 @@ import org.jetbrains.kotlin.name.NameUtils
 import org.jetbrains.kotlin.samWithReceiver.CliSamWithReceiverComponentContributor
 
 import org.jetbrains.kotlin.script.KotlinScriptDefinition
+import org.jetbrains.kotlin.scripting.compiler.plugin.ScriptingCompilerConfigurationComponentRegistrar
 
 import org.jetbrains.kotlin.utils.PathUtil
 import org.jetbrains.kotlin.utils.addToStdlib.firstNotNullResult
@@ -72,6 +74,24 @@ import java.io.OutputStream
 import java.io.PrintStream
 
 
+fun compileKotlinScriptModuleTo(
+    outputDirectory: File,
+    moduleName: String,
+    scriptFiles: Collection<String>,
+    scriptDef: KotlinScriptDefinition,
+    classPath: Iterable<File>,
+    logger: Logger,
+    pathTranslation: (String) -> String
+) = compileKotlinScriptModuleTo(
+    outputDirectory,
+    moduleName,
+    scriptFiles,
+    scriptDef,
+    classPath,
+    LoggingMessageCollector(logger, pathTranslation)
+)
+
+
 internal
 fun compileKotlinScriptToDirectory(
     outputDirectory: File,
@@ -79,19 +99,42 @@ fun compileKotlinScriptToDirectory(
     scriptDef: KotlinScriptDefinition,
     classPath: List<File>,
     messageCollector: LoggingMessageCollector
-): String =
+): String {
 
+    compileKotlinScriptModuleTo(
+        outputDirectory,
+        "buildscript",
+        listOf(scriptFile.path),
+        scriptDef,
+        classPath,
+        messageCollector
+    )
+
+    return NameUtils.getScriptNameForFile(scriptFile.name).asString()
+}
+
+
+private
+fun compileKotlinScriptModuleTo(
+    outputDirectory: File,
+    moduleName: String,
+    scriptFiles: Collection<String>,
+    scriptDef: KotlinScriptDefinition,
+    classPath: Iterable<File>,
+    messageCollector: LoggingMessageCollector
+) {
     withRootDisposable {
 
         withCompilationExceptionHandler(messageCollector) {
 
             val configuration = compilerConfigurationFor(messageCollector).apply {
-                addKotlinSourceRoot(scriptFile.canonicalPath)
                 put(JVM_TARGET, JVM_1_8)
                 put(RETAIN_OUTPUT_IN_MEMORY, false)
                 put(OUTPUT_DIRECTORY, outputDirectory)
-                setModuleName("buildscript")
+                setModuleName(moduleName)
+                addScriptingCompilerComponents()
                 addScriptDefinition(scriptDef)
+                scriptFiles.forEach { addKotlinSourceRoot(it) }
                 classPath.forEach { addJvmClasspathRoot(it) }
             }
             val environment = kotlinCoreEnvironmentFor(configuration).apply {
@@ -100,10 +143,9 @@ fun compileKotlinScriptToDirectory(
 
             compileBunchOfSources(environment)
                 || throw ScriptCompilationException(messageCollector.errors)
-
-            NameUtils.getScriptNameForFile(scriptFile.name).asString()
         }
     }
+}
 
 
 private
@@ -122,6 +164,7 @@ object HasImplicitReceiverCompilerPlugin {
 internal
 fun compileToDirectory(
     outputDirectory: File,
+    moduleName: String,
     sourceFiles: Iterable<File>,
     logger: Logger,
     classPath: Iterable<File>
@@ -132,7 +175,7 @@ fun compileToDirectory(
             val configuration = compilerConfigurationFor(messageCollector).apply {
                 addKotlinSourceRoots(sourceFiles.map { it.canonicalPath })
                 put(OUTPUT_DIRECTORY, outputDirectory)
-                setModuleName(outputDirectory.nameWithoutExtension)
+                setModuleName(moduleName)
                 classPath.forEach { addJvmClasspathRoot(it) }
                 addJvmClasspathRoot(kotlinStdlibJar)
             }
@@ -255,6 +298,7 @@ private
 fun compilerConfigurationFor(messageCollector: MessageCollector): CompilerConfiguration =
     CompilerConfiguration().apply {
         put(CLIConfigurationKeys.MESSAGE_COLLECTOR_KEY, messageCollector)
+        put(JVMConfigurationKeys.USE_FAST_CLASS_FILES_READING, true)
         put(CommonConfigurationKeys.LANGUAGE_VERSION_SETTINGS, gradleKotlinDslLanguageVersionSettings)
     }
 
@@ -279,6 +323,15 @@ fun CompilerConfiguration.setModuleName(name: String) {
 }
 
 
+private
+fun CompilerConfiguration.addScriptingCompilerComponents() {
+    add(
+        ComponentRegistrar.PLUGIN_COMPONENT_REGISTRARS,
+        ScriptingCompilerConfigurationComponentRegistrar()
+    )
+}
+
+
 private
 fun CompilerConfiguration.addScriptDefinition(scriptDef: KotlinScriptDefinition) {
     add(JVMConfigurationKeys.SCRIPT_DEFINITIONS, scriptDef)
